// Hopfield Neural Network
// Written by David J. Stein, Esq. See "License.html" for terms and conditions of use.
//
// This module contains the code for the Hopfield neural network.

#region Using declarations

using System;
using System.Collections;

#endregion

namespace Hopfield_Network_NS {

	internal class Neuron {

		// This class represents a neuron. It's really just a data structure, containing a single element for whether or not it is currently activated, and an array of synapses.

		#region Variables and initialization code

		internal bool x;
		internal ArrayList aSynapses;

		public Neuron() {
			aSynapses = new ArrayList();
		}

		#endregion

	}

	internal class Synapse {
		// This class represents a synapse between two neurons.

		#region Variables and initialization code

		internal Neuron[] neurons;
		internal int weight;
		internal Synapse(Neuron neuron1, Neuron neuron2) {
			neurons = new Neuron[2];
			neurons[0] = neuron1;
			neurons[1] = neuron2;
			weight = 0;
		}

		#endregion

	}

	internal class HopfieldNetwork {
		// This is the central class for a Hopfield network.

		#region Variables and initialization code

		internal string strName;  // the name of the neural network
		internal int iWidth, iHeight;  // the dimensions of the neural network
		internal int iTemplates;  // the total number of templates
		internal ArrayList aNeurons;  // an array of neurons
		internal ArrayList aSynapses;   // an array of synapses
		internal bool[] bInputPattern;  // the current input pattern

		internal HopfieldNetwork(int iWidth, int iHeight) {
			strName = "Sample Network";
			CreateNetwork(iWidth, iHeight);
		}

		#endregion

		#region Main functions

		internal void CreateNetwork(int iWidth, int iHeight) {
			// This class creates a new, empty network of specified size, with all neurons wired together.
			this.iWidth = iWidth;
			this.iHeight = iHeight;
			iTemplates = 0;
			// create neurons
			aNeurons = new ArrayList();
			for (int i = 0; i < iWidth * iHeight; i++)
				aNeurons.Add(new Neuron());
			// create synapses
			aSynapses = new ArrayList();
			foreach (Neuron neuron1 in aNeurons) {
				foreach (Neuron neuron2 in aNeurons) {
					if (neuron1 != neuron2) {
						Synapse synapse = new Synapse(neuron1, neuron2);
						aSynapses.Add(synapse);
						neuron1.aSynapses.Add(synapse);
						neuron2.aSynapses.Add(synapse);
					}
				}
			}
			// create array for input pattern
			bInputPattern = new bool[iWidth * iHeight];
		}

		internal void AddTemplate() {
			// This function stores a training template within the neuron synapse information.
			// put input pattern on inputs
			for (int i = 0; i < bInputPattern.Length; i++)
				((Neuron) aNeurons[i]).x = bInputPattern[i];
			// adjust weights
			foreach (object synapseobject in aSynapses) {
				Synapse s = (Synapse) synapseobject;
				if (s.neurons[0] != s.neurons[1])  // only adjust weights for synapses between two different neurons
					s.weight += ((s.neurons[0].x == s.neurons[1].x) ? +1 : -1);
			}
			// increase template count
			iTemplates++;
		}

		internal int Analyze(HopfieldNetworkApplication a) {
			// This function is the heart of the recognition process. It processes each neuron (once) based on the current activity and synapse relationship  of every other neuron. The neurons are updated in random order so that patterns are not recognized with different exactitude based on edge details, etc.
			int iCount = 0;
			int iNeuron;
			int iNet;
			int iTotalChange = 0;
			Neuron neuron;
			Synapse synapse;
			bool[] updated = new bool[aNeurons.Count];
			for (int i = 0; i < updated.Length; i++)
				updated[i] = false;
			Random r = new Random();
			// keep choosing neurons at random, and update those that aren't updated yet, until all neurons have been updated once
			while (iCount < aNeurons.Count) {
				iNeuron = r.Next(aNeurons.Count);
				if (updated[iNeuron] == false) {
					updated[iNeuron] = true;
					neuron = (Neuron) aNeurons[iNeuron];
					iCount++;
					iNet = 0;
					foreach (object synapseobject in aSynapses) {
						synapse = (Synapse) synapseobject;
						if (synapse.neurons[0] == neuron)
							iNet += synapse.weight * (((Neuron) synapse.neurons[1]).x == true ? +1 : -1);
						else if (synapse.neurons[1] == neuron)
							iNet += synapse.weight * (((Neuron) synapse.neurons[0]).x == true ? +1 : -1);
					}
					if (iNet > 0) {
						if (neuron.x == false)
							iTotalChange++;
						neuron.x = true;
					}
					else if (iNet < 0) {
						if (neuron.x == true)
							iTotalChange++;
						neuron.x = false;
					}
					else
						neuron.x = neuron.x;
				}
				a.picOutputPattern.Refresh();
			}
			return iTotalChange;
		}

		#endregion

	}
}
